from typing import Any, Callable, Generator, Generic, NoReturn, Optional, ParamSpec, TypedDict, TypeVar, Type, TypeAlias

# TODO I think we can simplify this just like we're doing with canisters
from .compiler.custom_modules.principal import Principal

Principal = Principal

int64 = int
int32 = int
int16 = int
int8 = int

nat = int
nat64 = int
nat32 = int
nat16 = int
nat8 = int

float64 = float
float32 = float

text = str

T = TypeVar('T')
opt = Optional[T]

Record = TypedDict
Variant = TypedDict

blob = bytes

null: TypeAlias = None

reserved = Any
empty = NoReturn # TODO in Python 3.11 I believe there is a Never type

Async = Generator[Any, Any, T]

def query(func: object):
    return func

def update(func: object):
    return func

def canister(cls: T) -> T:
    return cls

def init(func: object):
    return func

def heartbeat(func: object):
    return func

def pre_upgrade(func: object):
    return func

def post_upgrade(func: object):
    return func

def inspect_message(func: object):
    return func

# TODO need service

Query = Callable
Update = Callable
Oneway = Callable

class CanisterResult(Generic[T]):
    ok: T
    err: Optional[str]

    def __init__(self, ok: T, err: str):
        self.ok = ok
        self.err = err

    def notify(self) -> "NotifyResult": ...

    def with_cycles(self, cycles: nat64) -> "CanisterResult[T]": ...

    def with_cycles128(self, cycles: nat) -> "CanisterResult[T]": ...

# TODO Once RustPython supports Python 3.11, we can use the below and unify CanisterResult with the other Variants
# TODO The problem is that you can't really use generics with TypedDict yet: https://github.com/python/cpython/issues/89026
# TODO We could also consider a hack where we remove all references to CanisterResult before runtime, since this is really an analysis-time consideration
# class CanisterResult(Variant, Generic[T], total=False):
#     ok: T
#     err: str

class RejectionCode(Variant, total=False):
    NoError: None
    SysFatal: None
    SysTransient: None
    DestinationInvalid: None
    CanisterReject: None
    CanisterError: None
    Unknown: None

# TODO we might want this to act more like CanisterResult
class NotifyResult(Variant, total=False):
    ok: None
    err: RejectionCode

# TODO remove temp on StableMemoryError, StableGrowResult, Stable64GrowResult once multiple file analysis works
class StableMemoryErrorTemp(Variant, total=False):
    OutOfMemory: None
    OutOfBounds: None

class StableGrowResultTemp(Variant, total=False):
    ok: nat32
    err: StableMemoryErrorTemp

class Stable64GrowResultTemp(Variant, total=False):
    ok: nat64
    err: StableMemoryErrorTemp

# TODO we might want to just make this a dict[int, str] to keep the derive simple, or maybe a class with a principal and method name field
def Func(callable: Callable) -> Type[tuple[Principal, str]]: # type: ignore
    return type((Principal.from_str('aaaaa-aa'), ''))

# TODO make sure call_raw with notify works
class ic(Generic[T]):
    @staticmethod
    def call_raw(canister_id: Principal, method: str, args_raw: blob, payment: nat64) -> CanisterResult[T]:
        return AsyncInfo('call_raw', [
            canister_id,
            method,
            args_raw,
            payment
        ]) # type: ignore

    @staticmethod
    def call_raw128(canister_id: Principal, method: str, args_raw: blob, payment: nat) -> CanisterResult[T]:
        return AsyncInfo('call_raw128', [
            canister_id,
            method,
            args_raw,
            payment
        ]) # type: ignore

    @staticmethod
    def accept_message():
        _kybra_ic.accept_message() #type:ignore

    @staticmethod
    def candid_encode(candid_string: str) -> blob:
        return _kybra_ic.candid_encode(candid_string) # type: ignore

    @staticmethod
    def candid_decode(candid_encoded: blob) -> str:
        return _kybra_ic.candid_decode(candid_encoded) # type: ignore

    @staticmethod
    def canister_balance() -> nat64:
        return _kybra_ic.canister_balance() # type: ignore

    @staticmethod
    def canister_balance128() -> nat:
        return _kybra_ic.canister_balance128() # type: ignore

    @staticmethod
    def method_name() -> str:
        return _kybra_ic.method_name() #type:ignore

    @staticmethod
    def msg_cycles_accept(max_amount: nat64) -> nat64:
        return _kybra_ic.msg_cycles_accept(max_amount) # type: ignore

    @staticmethod
    def msg_cycles_accept128(max_amount: nat) -> nat:
        return _kybra_ic.msg_cycles_accept128(max_amount) # type: ignore

    @staticmethod
    def msg_cycles_available() -> nat64:
        return _kybra_ic.msg_cycles_available() # type: ignore

    @staticmethod
    def msg_cycles_available128() -> nat:
        return _kybra_ic.msg_cycles_available128() # type: ignore

    @staticmethod
    def msg_cycles_refunded() -> nat64:
        return _kybra_ic.msg_cycles_refunded() # type: ignore

    @staticmethod
    def msg_cycles_refunded128() -> nat:
        return _kybra_ic.msg_cycles_refunded128() # type: ignore

    @staticmethod
    def notify_raw(
        canister_id: Principal,
        method: str,
        args_raw: blob,
        payment: nat
    ) -> NotifyResult:
        return _kybra_ic.notify_raw(canister_id, method, args_raw, payment) # type: ignore

    @staticmethod
    def print(x: Any):
        _kybra_ic.print(str(x)) # type: ignore

    @staticmethod
    def stable_bytes() -> blob:
        return _kybra_ic.stable_bytes() # type: ignore

    @staticmethod
    def stable_grow(new_pages: nat32) -> StableGrowResultTemp:
        return _kybra_ic.stable_grow(new_pages) # type: ignore

    @staticmethod
    def stable_read(offset: nat32, length: nat32) -> blob:
        return _kybra_ic.stable_read(offset, length) # type: ignore

    @staticmethod
    def stable_size() -> nat32:
        return _kybra_ic.stable_size() # type: ignore

    @staticmethod
    def stable_write(offset: nat32, buf: blob):
        _kybra_ic.stable_write(offset, buf) # type: ignore

    @staticmethod
    def stable64_grow(new_pages: nat64) -> Stable64GrowResultTemp:
        return _kybra_ic.stable64_grow(new_pages) # type: ignore

    @staticmethod
    def stable64_read(offset: nat64, length: nat64) -> blob:
        return _kybra_ic.stable64_read(offset, length) # type: ignore

    @staticmethod
    def stable64_size() -> nat64:
        return _kybra_ic.stable64_size() # type: ignore

    @staticmethod
    def stable64_write(offset: nat64, buf: blob):
        _kybra_ic.stable64_write(offset, buf) # type: ignore

    @staticmethod
    def trap(message: str) -> NoReturn: # type: ignore
        _kybra_ic.trap(message) # type: ignore

class Canister:
    canister_id: Principal

    def __init__(self, canister_id: Principal):
        self.canister_id = canister_id

P = ParamSpec('P')

class AsyncInfo:
    name: str
    args: list[Any]

    def __init__(self, name: str, args: list[Any]):
        self.name = name
        self.args = args

    def with_cycles(self, cycles: nat64) -> "AsyncInfo":
        return AsyncInfo('call_with_payment', [*self.args, cycles])

    def with_cycles128(self, cycles: nat) -> "AsyncInfo":
        return AsyncInfo('call_with_payment128', [*self.args, cycles])

    def notify(self) -> NotifyResult:
        # TODO calculate the notify function name here...actually, maybe we should just do this in the same way as the other calls? Just to keep it simple?
        return _kybra_ic['notify_function_name'] # type: ignore

# TODO this decorator is removing the static type checking of the self parameter for instance methods
# TODO watch out for *kwargs
def method(func: Callable[P, T]) -> Callable[P, CanisterResult[T]]:
    def intermediate_func(*args): # type: ignore
        the_self = args[0] # type: ignore
        selfless_args = args[1:] # type: ignore

        return AsyncInfo('call', [
            the_self.canister_id, # type: ignore
            func.__qualname__,
            *selfless_args
        ])

    return intermediate_func # type: ignore



from kybra import blob, Canister, method, Principal
from kybra.canisters.management.basic import CanisterStatusResult, DepositCyclesArgs, DeleteCanisterArgs, CreateCanisterArgs, CreateCanisterResult, InstallCodeArgs, UninstallCodeArgs, UpdateSettingsArgs, StartCanisterArgs, StopCanisterArgs, CanisterStatusArgs, ProvisionalCreateCanisterWithCyclesArgs, ProvisionalCreateCanisterWithCyclesResult, ProvisionalTopUpCanisterArgs

# TODO change the return types to void

class ManagementCanister(Canister):
    @method
    def create_canister(self, args: CreateCanisterArgs) -> CreateCanisterResult: ...

    @method
    def update_settings(self, args: UpdateSettingsArgs) -> None: ...

    @method
    def install_code(self, args: InstallCodeArgs) -> None: ...

    @method
    def uninstall_code(self, args: UninstallCodeArgs) -> None: ...

    @method
    def start_canister(self, args: StartCanisterArgs) -> None: ...

    @method
    def stop_canister(self, args: StopCanisterArgs) -> None: ...

    @method
    def canister_status(self, args: CanisterStatusArgs) -> CanisterStatusResult: ...

    @method
    def delete_canister(self, args: DeleteCanisterArgs) -> None: ...

    @method
    def deposit_cycles(self, args: DepositCyclesArgs) -> None: ...

    @method
    def raw_rand(self) -> blob: ...

    @method
    def provisional_create_canister_with_cycles(self, args: ProvisionalCreateCanisterWithCyclesArgs) -> ProvisionalCreateCanisterWithCyclesResult: ...

    @method
    def provisional_top_up_canister(self, args: ProvisionalTopUpCanisterArgs) -> None: ...

management_canister = ManagementCanister(Principal.from_str('aaaaa-aa'))



# type: ignore

import kybra
import modulegraph.modulegraph
import os
import shutil
import sys
from pathlib import Path

# This is the name of the canister passed into python -m kybra from the dfx.json build command
canister_name = sys.argv[1]

print(f'\nBuilding canister {canister_name}\n')

# This is the path to the developer's entry point Python file passed into python -m kybra from the dfx.json build command
py_entry_file_path = sys.argv[2]

# This is the Python module name of the developer's Python project, derived from the entry point Python file passed into python -m kybra from the dfx.json build command
py_entry_module_name = Path(py_entry_file_path).stem

# This is the location of all code used to generate the final canister Rust code
canister_path=f'.dfx/kybra/{canister_name}'

py_file_names_file_path = f'{canister_path}/file_names.txt'

# This is the path to the developer's Candid file passed into python -m kybra from the dfx.json build command
did_path = sys.argv[3]

# This is the path to the Kybra compiler Rust code delivered with the Python package
compiler_path = os.path.dirname(kybra.__file__) + '/compiler'

# This is the final generated Rust file that is the canister
lib_path=f'{canister_path}/src/lib.rs'

# This is the location of the Candid file generated from the final generated Rust file
generated_did_path = f'{canister_path}/index.did'

# This is the Rust target directory
target_path=f'{canister_path}/target'

cargo_bin_root = os.environ.get('CARGO_INSTALL_ROOT') or os.environ.get('CARGO_HOME') or f'{os.environ["HOME"]}/.cargo'

# This is where we store custom Python modules, such as stripped-down versions of stdlib modules
custom_modules_path = f'{compiler_path}/custom_modules'

# Copy all of the Rust project structure from the pip package to an area designed for Rust compiling
shutil.copytree(compiler_path, canister_path, dirs_exist_ok=True)

# Begin module bundling/gathering process
path = list(filter(lambda x: x.startswith(os.getcwd()), sys.path)) + [os.path.dirname(py_entry_file_path)]

graph = modulegraph.modulegraph.ModuleGraph(path)
entry_point = graph.run_script(py_entry_file_path)

# We want to bundle/gather all Python files into the python_source directory for RustPython freezing
# The location that Kybra will look to when running py_freeze!
# py_freeze! will compile all of the Python code in the directory recursively (modules must have an __init__.py to be included)
python_source_path = f'{canister_path}/python_source'

if os.path.exists(python_source_path):
    shutil.rmtree(python_source_path)

os.makedirs(python_source_path)

# Copy our custom Python modules into the python_source directory
shutil.copytree(custom_modules_path, python_source_path, dirs_exist_ok=True)

flattened_graph = list(graph.flatten(start=entry_point))

for node in flattened_graph:
    if type(node) == modulegraph.modulegraph.Script:
        shutil.copy(node.filename, f'{python_source_path}/{os.path.basename(node.filename)}')

    if type(node) == modulegraph.modulegraph.SourceModule:
        shutil.copy(node.filename, f'{python_source_path}/{os.path.basename(node.filename)}')

    if type(node) == modulegraph.modulegraph.Package:
        shutil.copytree(node.packagepath[0], f'{python_source_path}/{node.identifier}', dirs_exist_ok=True)

    if type(node) == modulegraph.modulegraph.NamespacePackage:
        shutil.copytree(node.packagepath[0], f'{python_source_path}/{node.identifier}', dirs_exist_ok=True)

py_file_names = list(
    filter(
        lambda filename: filename is not None,
        map(
            lambda node: node.filename,
            filter(
                lambda node: node.filename is not '-', # This filters out namespace packages
                flattened_graph
            )
        )
    )
)

# Rust prerequisites
os.system('rustup target add wasm32-unknown-unknown')
# TODO this should eventually be replaced with ic-wasm once this is resolved: https://forum.dfinity.org/t/wasm-module-contains-a-function-that-is-too-complex/15407/43?u=lastmjs
os.system('cargo install ic-cdk-optimizer --version 0.3.4 || true')
# os.system('cargo install ic-wasm --version 0.3.0 || true')

py_file_names_file = open(py_file_names_file_path, 'w')
py_file_names_file.write(','.join(py_file_names))
py_file_names_file.close()

# Generate the Rust code
print('[1/3] 🔨 Compiling Python...')
os.system(f'CARGO_TARGET_DIR={target_path} cargo run --manifest-path {canister_path}/kybra_generate/Cargo.toml {py_file_names_file_path} {py_entry_module_name} | rustfmt --edition 2018 > {lib_path}')

# Compile the generated Rust code
print('[2/3] 🚧 Building Wasm binary...')
os.system(f'CARGO_TARGET_DIR={target_path} cargo build --manifest-path {canister_path}/Cargo.toml --target wasm32-unknown-unknown --package kybra_generated_canister --release')

# Generate the Candid file
print('[3/3] 📝 Generating Candid file...')
os.system(f'CARGO_TARGET_DIR={target_path} cargo test --manifest-path {canister_path}/Cargo.toml')

# Copy the generated Candid file to the developer's source directory
os.system(f'cp {generated_did_path} {did_path}')

# Optimize the Wasm binary
# TODO this should eventually be replaced with ic-wasm once this is resolved: https://forum.dfinity.org/t/wasm-module-contains-a-function-that-is-too-complex/15407/43?u=lastmjs
os.system(f'{cargo_bin_root}/bin/ic-cdk-optimizer {target_path}/wasm32-unknown-unknown/release/kybra_generated_canister.wasm -o {target_path}/wasm32-unknown-unknown/release/{canister_name}.wasm')
# os.system(f'{cargo_bin_root}/bin/ic-wasm {target_path}/wasm32-unknown-unknown/release/kybra_generated_canister.wasm -o {target_path}/wasm32-unknown-unknown/release/{canister_name}.wasm shrink')

# gzip the Wasm binary
os.system(f'gzip -f -k {target_path}/wasm32-unknown-unknown/release/{canister_name}.wasm')



from kybra import blob, nat, opt, Principal, Record, Variant

# TODO type aliases do not work yet
# TODO many canister_id fields need to be changed to use this alias
# CanisterId = Principal
# WasmModule = blob

class CreateCanisterArgs(Record):
    settings: opt["CanisterSettings"]

class CanisterSettings(Record):
    controllers: opt[list[Principal]]
    compute_allocation: opt[nat]
    memory_allocation: opt[nat]
    freezing_threshold: opt[nat]

class DefiniteCanisterSettings(Record):
    controllers: list[Principal]
    compute_allocation: nat
    memory_allocation: nat
    freezing_threshold: nat

class CreateCanisterResult(Record):
    canister_id: Principal

class UpdateSettingsArgs(Record):
    canister_id: Principal
    settings: CanisterSettings

# TODO this should be below InstallCodeArgs once double quote type syntax is allowed
class InstallCodeMode(Variant, total=False):
    install: None
    reinstall: None
    upgrade: None

class InstallCodeArgs(Record):
    mode: InstallCodeMode
    canister_id: Principal
    wasm_module: blob
    arg: blob

class UninstallCodeArgs(Record):
    canister_id: Principal

class StartCanisterArgs(Record):
    canister_id: Principal

class StopCanisterArgs(Record):
    canister_id: Principal

class CanisterStatusArgs(Record):
    canister_id: Principal

# TODO this should be below CanisterStatusResult once double quote type syntax is allowed
class CanisterStatus(Variant):
    running: None
    stopping: None
    stopped: None

class CanisterStatusResult(Record):
    status: CanisterStatus
    settings: DefiniteCanisterSettings
    module_hash: opt[blob]
    memory_size: nat
    cycles: nat

class DeleteCanisterArgs(Record):
    canister_id: Principal

class DepositCyclesArgs(Record):
    canister_id: Principal

class ProvisionalCreateCanisterWithCyclesArgs(Record):
    amount: opt[nat]
    settings: opt[CanisterSettings]

class ProvisionalCreateCanisterWithCyclesResult(Record):
    canister_id: Principal

class ProvisionalTopUpCanisterArgs(Record):
    canister_id: Principal
    amount: nat



from setuptools import setup

setup(
    name='kybra',
    version='0.0.7',
    package_data={
        'kybra': ['compiler/**', 'canisters/**']
    },
    include_package_data=True,
    packages=['kybra'],
    install_requires=['modulegraph==0.19.3']
)
